{-----------------------------------------------------------------------------
 Package:        MD - CP Gateway Service
 Developer(s):   Brian Juergensmeyer, Dan Petit
 Date Created:   13-Oct-2006
 Site Name:      Hines OIFO
 Description:    This is the primary CP Gateway service file.  It contains all
                 code required to start, stop, pause, and restart the service.
                 It also contains the event handler used by TCPNotifier to notify
                 the service that an incoming message is waiting for processing.
                 The event handler will then contact VistA via the RPC Broker and
                 retrieve the value of the message.
-----------------------------------------------------------------------------}

unit sCPGateway;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  SvcMgr,
  Dialogs,
  ExtCtrls,
  TRPCB,
  ActnList,
  oCPGatewayConfig,
  oCPParameters,
  uCPNotifier,
  uCPMessage,
  oHL7Message,
  oInterfaces,
  oClioFilerObs,
  XMLDoc,
  XMLIntf,
  XMLDom,
  Contnrs,
  oClio,
  oGateway;

const
  CONTEXT = 'MDCP GATEWAY CONTEXT';
  // Broker Calls
  RPC_GET_CORRECTIONS = 'MDCP CORRECTIONS BY IEN';
  RPC_GET_MESSAGE_DETAIL = 'MDCP MESSAGE BY IEN';
  RPC_GET_MESSAGE_LIST = 'MDCP RESULTS BY STATUS';
  RPC_UPDATE_STATUS = 'MDCP UPDATE MESSAGE STATUS';
  RPC_UPDATE_REASON = 'MDCP UPDATE MESSAGE REASON';

type
  TQualifierType = (qtTerm, qtUOM, qtMethod, qtQuality, qtProduct);

  TCPGatewaySvc = class(TService, ICPGatewayLogger)
    acList: TActionList;
    acPoller: TAction;
    acRestartBroker: TAction;
    acConnectBroker: TAction;
    acDisconnectBroker: TAction;
    acExecuteTimer: TAction;
    acPollForMessages: TAction;
    tmrOneTimeLoad: TTimer;
    acAddServiceDescriptionToRegistry: TAction;
    procedure acRestartBrokerExecute(Sender: TObject);
    procedure acConnectBrokerExecute(Sender: TObject);
    procedure acDisconnectBrokerExecute(Sender: TObject);
    procedure acPollForMessagesExecute(Sender: TObject);
    procedure ServiceExecute(Sender: TService);
    procedure ServicePause(Sender: TService; var Paused: Boolean);
    procedure ServiceContinue(Sender: TService; var Continued: Boolean);
    procedure ServiceShutdown(Sender: TService);
    procedure ServiceStart(Sender: TService; var Started: Boolean);
    procedure ServiceStop(Sender: TService; var Stopped: Boolean);
    procedure ServiceCreate(Sender: TObject);
    procedure ServiceDestroy(Sender: TObject);
    procedure tmrOneTimeLoadTimer(Sender: TObject);
    procedure acAddServiceDescriptionToRegistryExecute(Sender: TObject);
  private
    fCPParameters: ICPParameters;
    FConfiguration: ICPGatewayConfig;
    FDBConnection: IDBConnection;
    FTerminologySource: ITerminology;
    FCliOService: ICliOService;
    FObservationFiler: ICliOFiler;
    FAlarmList: ICliOAlarmList;
    FBroker: ILegacyBroker;
    FNotifier: TCPNotifier;
    FLogLevel: Integer;
    FMappingErrorList: TStringList;
    procedure FNotifierOnNotify(Sender: TObject; AMessage: TCPMessage);
    function GetMessage(IEN: string): string;
    procedure ChangeMessageStatus(AMessage: TCPHL7Message; const Status: string; const StatusReason: TStringList = nil);
    procedure Cleanup;
  public
    function GenHL7_XMLMessage(AHL7Message: TCPHL7Message): ICliOFilerDoc;
    function GetServiceController: TServiceController; override;
    //procedure Log(const AMessage: string; const AMessageType: word; const ALogLevel: TLogLevel);
    procedure Log(const AMessage: string; const ALogLevel: TLogLevel); stdcall;
    { Public declarations }
  end;

var
  CPGatewaySvc: TCPGatewaySvc;

implementation

uses
  SyncObjs,
  ActiveX,
  Masks,
  MFunStr,
  oClioTypes,
  uHL7CorrectionMap,
  Variants,
  uParser,
  DateUtils,
  oTermMapper,
  Registry,
  Hash;

var
  CritSec: TCriticalSection;

const
  PENDING_GATEWAY_ACTION = '2';
  PROCESSED = '4';
  ERROR = '3';
  SVC_VERSION = '1.0.16.270';

  //XML document tags.
  CLIO = 'CLIO';
  CLIO_TYPE = 'TYPE';
  CLIO_VERSION = 'VERSION';
  BATTERY = 'BATTERY';
  BATTERY_ID = 'BATTERY_ID';
  FACILITY_ID = 'FACILITY_ID';
  PATIENT_ID = 'PATIENT_ID';
  OBSERVED_DATE_TIME = 'OBSERVED_DATE_TIME';
  OBSERVED_BY_ID = 'OBSERVED_BY_ID';
  ENTERED_DATE_TIME = 'ENTERED_DATE_TIME';
  ENTERED_BY_ID = 'ENTERED_BY_ID';
  FACILITY_LOCATION_ID = 'FACILITY_LOCATION_ID';
  COMMENTS = 'COMMENTS';
  OBSERVATIONS = 'OBSERVATIONS';
  OBSERVATION = 'OBSERVATION';
  OBS_ID = 'OBS_ID';
  PARENT_ID = 'PARENT_ID';
  QUALIFIERS = 'QUALIFIERS';
  QUALIFIER = 'QUALIFIER';
  QUALIFIER_NAME = 'NAME';
  QUALIFIER_TYPE = 'TYPE';
  QUALIFIER_VUID = 'VUID';
  SVALUE = 'SVALUE';
  HL7_SOURCE = 'SOURCE';
  HL7_SOURCE_COMMENTS = 'SOURCE_COMMENTS';
  HL7_SOURCE_DATA_ITEM_ID = 'SOURCE_DATA_ITEM_ID';
  HL7_SOURCE_VERSION = 'SOURCE_VERSION';
  TERM_ID = 'TERM_ID';
  UNIT_ID = 'UNIT_ID';
  METHOD_ID = 'METHOD_ID';
  LOCATION_ID = 'LOCATION_ID';
  POSITION_ID = 'POSITION_ID';
  QUALITY_ID = 'QUALITY_ID';
  PRODUCT_ID = 'PRODUCT_ID';
  COMMENT_ID = 'COMMENT_ID';
  DATE_TIME_FORMAT = 'YYYY-MM-DD HH:MM:SS';

  //Registry information.
  REG_KEY = '\SOFTWARE\Vista\Clinical Procedures\CPGateway';
  REG_DESC_KEY = '\System\CurrentControlSet\Services\CPGatewaySvc';

  MTH = 'MTH';
  LOC = 'LOC';
  POS = 'POS';
  QUA = 'QUA';
  PRO = 'PRO';
  CMT = 'CMT';
  OBR = 'OBR';
  OBX = 'OBX';

  DO_NOT_PROCESS = '{4615254C-EC67-46D0-BF70-9A54BEB4B32D}';

{$R *.DFM}

function GenCGUID: string;
var
  GUID: TGUID;
begin
  CoCreateGUID(GUID);
  result := GUIDToString(GUID);
  result := StringReplace(result, '-', '', [rfReplaceAll]);
  result := StringReplace(result, '{', '', [rfReplaceAll]);
  result := StringReplace(result, '}', '', [rfReplaceAll]);
end;

function GetSource(AField: string): string;
begin
  result := piece(piece(AField, '^', 2), '~', 1);
end;

function HL7DateTimeToTDateTime(AnHL7DateTime: string): TDateTime;
var
  AYear: word;
  AMonth: word;
  ADate: word;
  AnHour: word;
  AMinute: word;
  ASecond: word;
begin
  AYear := StrToIntDef(Copy(AnHL7DateTime, 1, 4), 0);
  AMonth := StrToIntDef(Copy(AnHL7DateTime, 5, 2), 0);
  ADate := StrToIntDef(Copy(AnHL7DateTime, 7, 2), 0);
  AnHour := StrToIntDef(Copy(AnHL7DateTime, 9, 2), 0);
  AMinute := StrToIntDef(Copy(AnHL7DateTime, 11, 2), 0);
  ASecond := StrToIntDef(Copy(AnHL7DateTime, 13, 2), 0);
  result := EncodeDateTime(AYear, AMonth, ADate, AnHour, AMinute, ASecond, 0);
end;

procedure ParseSourceData(AData: string; ANode: IXMLNode); overload;
var
  AList: TStringList;
  I: Integer;
begin
  AList := TStringList.Create;
  try
    AList.Delimiter := '~';
    AList.DelimitedText := StringReplace(AData, '^', '=', [rfReplaceAll]);

    for I := 0 to AList.Count - 1 do
      begin
        ANode.AddChild(AList.Names[i]).NodeValue := AList.ValueFromIndex[I];
      end;
    ANode.ChildNodes[HL7_SOURCE_COMMENTS].NodeValue := 'CPGatewayService.exe';
  finally
    FreeAndNil(AList);
  end;
end;

procedure ParseSourceData(AData: string; AList: TStringList); overload;
begin
  AList.Delimiter := '~';
  AList.DelimitedText := StringReplace(AData, '^', '=', [rfReplaceAll]);
end;

function GetObservationGroup(ASource: TStringList): TStringList;
var
  I: Integer;
  SegName: string;
  InOBRSeg: boolean;
begin
  result := TStringList.Create;
  I := ASource.Count - 1;
  InOBRSeg := false;
  while I > -1 do
    begin
      SegName := copy(ASource[I], 1, 3);
      if InObrSeg and (SegName = OBR) then
        exit;

      InOBRSeg := InOBRSeg or (SegName = OBR);

      if (SegName = OBR) or (SegName = OBX) then
        result.Insert(0, ASource[I]);

      ASource.Delete(I);
      dec(I);
    end;
end;

procedure ServiceController(CtrlCode: DWord); stdcall;
begin
  CPGatewaySvc.Controller(CtrlCode);
end;

{Public funtion}

function TCPGatewaySvc.GetServiceController: TServiceController;
begin
  Result := ServiceController;
end;

{Service methods}

procedure TCPGatewaySvc.ServiceExecute(Sender: TService);
begin
  while not Terminated do
    ServiceThread.ProcessRequests(True);
end;

procedure TCPGatewaySvc.ServicePause(Sender: TService; var Paused: Boolean);
begin
  acDisconnectBroker.Execute;
  LogMessage('CP Gateway Service Paused', EVENTLOG_INFORMATION_TYPE);
  Paused := True;
end;

procedure TCPGatewaySvc.ServiceContinue(Sender: TService; var Continued: Boolean);
begin
  try
    acConnectBroker.Execute;
    LogMessage('Service Resumed', EVENTLOG_INFORMATION_TYPE);
    Continued := True;
  except on
    E: Exception do
      begin
        LogMessage(Format('Service Not Resumed. Unable to establish RPC Broker connection: %s', [E.Message]), EVENTLOG_ERROR_TYPE);
        Continued := False;
      end;
  end;
end;

procedure TCPGatewaySvc.ServiceShutdown(Sender: TService);
begin
  try
    acDisconnectBroker.Execute;
  except
    LogMessage('Error in ServiceShutdown', EVENTLOG_ERROR_TYPE);
  end;
end;

procedure TCPGatewaySvc.ServiceStart(Sender: TService; var Started: Boolean);
begin
  CoInitialize(nil);

  FDBConnection := NewDBConnection_VistANoCCOW; {CCOW can really hose up a background service}
  Log('Trying to connect....', llCritical);
  Sleep(2000);
  acConnectBroker.Execute;
  Sleep(2000);
  Log('Back from acConnectBroker.Execute ....', llCritical);
  Sleep(2000);

  if not FDBConnection.IsConnected then
    begin
      Log('Unable to start service: no connection established.', llCritical);
      Started := False;
      Exit;
    end;

  FMappingErrorList := TStringList.Create;

  fConfiguration := NewCPGatewayConfig(FDBConnection);
  fConfiguration.Load;

  fCPParameters := NewCPParameters(FDBConnection);
  fCPParameters.Load;

  FNotifier := TCPNotifier.Create(fConfiguration, self as ICPGatewayLogger);
  FNotifier.OnCPNotify := FNotifierOnNotify;
  FNotifier.OnCPReload := tmrOneTimeLoadTimer;

  DefaultDOMVendor := 'MSXML';

  if fCPParameters.getApplicationRegistered then
    if Assigned(fConfiguration) then
      begin
        FLogLevel := FConfiguration.getLogLevel;

        if FDBConnection.QueryInterface(ILegacyBroker, FBroker) <> 0 then
          begin
            Log('Unable to start service: unable to get interface to legacy broker', llCritical);
          end;

        {Let the timer do the heavy lifing in 5 seconds so we can start up faster}
        tmrOneTimeLoad.Interval := 5000;
        tmrOneTimeLoad.Enabled := True;
        Log('Delayed terminology load scheduled', llDetail);

        Log('CP Gateway Service version ''' + fCPParameters.getCurrentApplicationVersion + ''' started', llImportant);
        acAddServiceDescriptionToRegistry.Execute;
        Started := True;
      end
    else
      begin
        Log('CP Gateway Service unable to start, no configuration found', llImportant);
        Started := False;
      end
  else
    begin
      Log('CP Gateway Service version ' + fCPParameters.getCurrentApplicationVersion +
        ' is unable to start, version ''' +
        fCPParameters.getRegisteredApplicationVersion + ''' is required', llCritical);
      Started := False;
    end;
  if not Started then
    Cleanup;
end;

procedure TCPGatewaySvc.ServiceStop(Sender: TService; var Stopped: Boolean);
begin
  try
    if fBroker.Connected then
      begin
        Log('Stopping Notifier', llDetail);
        FNotifier.Active := false;
        Log('Disconnecting Broker', llDetail);
        acDisconnectBroker.Execute;
      end;
    {Nil all IInterfaces}
    Cleanup;
    {
    fCPParameters := nil;
    FConfiguration := nil;
    FDBConnection := nil;
    FTerminologySource := nil;
    FCliOService := nil;
    FObservationFiler := nil;
    FBroker := nil;

    FreeAndNil(FNotifier);
    FreeAndNil(FMappingErrorList);
    }
    Log('CPGateway Service Stopped', llDetail);
    Stopped := True;
    CoUninitialize;
  except
    on E: Exception do
      Log('Error encounted stopping CPGateway Service ' + E.Message, llError);
  end;
end;

procedure TCPGatewaySvc.tmrOneTimeLoadTimer(Sender: TObject);
begin
  tmrOneTimeLoad.Enabled := False;

  if Assigned(FTerminologySource) then
    Log('Reloading Terminology', llImportant)
  else
    Log('Loading Terminology', llImportant);

  FTerminologySource := NewTerminology;
  FTerminologySource.Initialize(CONTROLLER_NAME[ctTerminology], FDBConnection);
  FTerminologySource.Load;
  Log('Terminology Loaded', llImportant);

  FCliOService := NewCliOService(FTerminologySource);
  FCliOService.Initialize('CLIO', FDBConnection);

  FObservationFiler := FCliOService.getNewFiler;
  FAlarmList := NewAlarmList(FDBConnection,FTerminologySource);
  FObservationFiler.SetAlarmList(FAlarmList);

  FNotifier.Port := FConfiguration.getGatewayServerPort;

  FNotifier.Active := true;

  Log('Terminology Post load Completed', llImportant);
end;

procedure TCPGatewaySvc.ServiceCreate(Sender: TObject);
begin
  CoInitialize(nil);
  CritSec := TCriticalSection.Create;
  DisplayName := 'CP Gateway Service';
end;

{Action List methods}

procedure TCPGatewaySvc.acRestartBrokerExecute(Sender: TObject);
begin
  Log('Trying RPC Broker reconnect', llError);
  FDBConnection.Disconnect;
  acConnectBroker.Execute;
end;

procedure TCPGatewaySvc.acAddServiceDescriptionToRegistryExecute(Sender: TObject);
var
  Reg: TRegistry;
begin
  Reg := TRegistry.Create(KEY_WRITE);
  try
    Reg.RootKey := HKEY_LOCAL_MACHINE;
    Reg.OpenKey(REG_DESC_KEY, False);

    Reg.WriteString('Description',
      Format('Clinical Procedures Gateway Service v%s - Compiled %s (CRC: %s)',
      [fCPParameters.getCurrentApplicationVersion, fCPParameters.getCurrentApplicationCompileDateTime, fCPParameters.getCurrentApplicationCRCValue]));
  finally
    FreeAndNil(Reg);
  end;
end;

procedure TCPGatewaySvc.acConnectBrokerExecute(Sender: TObject);
var
  Reg: TRegistry;
  ServerIP: string;
  Port: string;
  EncAccessCode: string;
  EncVerifyCode: string;
  ACode: string;
  VCode: string;
  aVistASystem: IVistASystem;
begin
  {Lets see if we can get a connection}

  try
    {
      Note: You can't use ICPGatewayConfig here, as you don't
      have a broker connection yet.
    }
    Reg := TRegistry.Create(KEY_READ);
    try
      Reg.RootKey := HKEY_LOCAL_MACHINE;
      Reg.OpenKeyReadOnly(REG_KEY);
      ServerIP := Reg.ReadString('SERVER');
      Port := IntToStr(Reg.ReadInteger('PORT'));
      EncAccessCode := Reg.ReadString('ACCESS CODE');
      EncVerifyCode := Reg.ReadString('VERIFY CODE');
      ACode := Hash.Decrypt(EncAccessCode);
      VCode := Hash.Decrypt(EncVerifyCode);
    finally
      FreeAndNil(Reg);
    end;

    LogMessage(Format('Attempting to connect the RPC Broker: Server %s Port %s', [ServerIP, Port]),
      EVENTLOG_INFORMATION_TYPE);

    if not FDBConnection.Connect([ServerIP, Port, ACode + ';' + VCode, CONTEXT]) then
      raise EBrokerError.Create('Not Sure Why');
    //LogMessage('Unable to connect to broker!', EVENTLOG_ERROR_TYPE);

    if FDBConnection.QueryInterface(IVistaSystem, aVistaSystem) = 0 then
      LogMessage(Format(
        'RPC Broker connection established: Server: %s, Port: %s, Account DUZ: %s',
        [ServerIP, Port, aVistaSystem.DUZ]), EVENTLOG_INFORMATION_TYPE)
    else
      LogMessage(Format(
        'RPC Broker connection established: %s',
        [fDBConnection.getDescription]), EVENTLOG_INFORMATION_TYPE);

    aVistASystem := nil;
  except
    on E: EBrokerError do
      begin
        LogMessage(Format(
          'RPC Broker connection failed: Server: %s, Port: %d, Error: %s',
          [fBroker.Server, fBroker.ListenerPort, E.Message]), EVENTLOG_ERROR_TYPE);
        acDisconnectBroker.Execute;
      end;
    on E: Exception do
      begin
        LogMessage(Format('General error occurred: %s', [E.Message]), EVENTLOG_ERROR_TYPE);
      end;
  end;
end;

procedure TCPGatewaySvc.acDisconnectBrokerExecute(Sender: TObject);
begin
  FDBConnection.Disconnect;
end;

procedure TCPGatewaySvc.ServiceDestroy(Sender: TObject);
begin
  FreeAndNil(CritSec);
  CoUninitialize;
end;

procedure TCPGatewaySvc.FNotifierOnNotify(Sender: TObject; AMessage: TCPMessage);
var
  Doc: ICliOFilerDoc;
  ErrorText: TStringList;
begin
  CoInitialize(nil);
  try
    CritSec.Enter;
    try
      {
        First, we need to get the rest of the HL7 message back out of VistA.
      }

      if not Assigned(AMessage) then
        begin
          Log('AMessage not assigned.', llCritical);
        end;

      try
        if AMessage is TCPHL7Message then
          Doc := GenHL7_XMLMessage(TCPHL7Message(AMessage))
        else
          begin
            {
              Retrieval of query messages will occur here.
            }
          end;

        if not Assigned(Doc) then
          raise Exception.Create('XMLMessage not assigned when trying to save to CliO');

        if FMappingErrorList.Count > 0 then
          begin {Report mapping errors on down to the CP_HL7_LOG_REASON}
            if AMessage is TCPHL7Message then
              ChangeMessageStatus(TCPHL7Message(AMessage), ERROR, FMappingErrorList);
          end
        else
          begin
            Log('Preparing to save document in CliO', llDetail);
            FObservationFiler.Execute(Doc);

            Log('Document is in CliO', llDetail);
            if AMessage is TCPHL7Message then
              ChangeMessageStatus(TCPHL7Message(AMessage), PROCESSED);

            Log(Format('Message ''%s'' marked as processed.', [aMessage.ID]), llImportant);
          end;
      except
        on E: Exception do
          begin
            ErrorText := TStringList.Create;
            try
              {
                Most likely, we failed validation.  So, we need to return an error to
                VistAland, so that we can mark
              }
              ErrorText.Text := FObservationFiler.getErrorLog;
              ErrorText.Insert(0, E.Message);
              if AMessage is TCPHL7Message then
                ChangeMessageStatus(TCPHL7Message(AMessage), ERROR, ErrorText);
            finally
              FreeAndNil(ErrorText);
            end;
          end;
      end;
      Doc := nil;
    finally
      CritSec.Leave;
    end;
  finally
    CoUninitialize;
  end;
end;

function TCPGatewaySvc.GetMessage(IEN: string): string;
var
  ShouldStop: Boolean;
begin
  if not FBroker.Connected then
    begin
      Log('Message received, but RPC Broker not connected.  Trying to connect to broker.', llImportant);
      acConnectBrokerExecute(nil);
      if not FBroker.Connected then
        begin
          Log('Message received and unable to connect to broker.  Stopping Service.', llCritical);
          ShouldStop := true;
          ServiceStop(self, ShouldStop);
        end;
    end;

  Log(Format('Retrieving message with ID: %s', [IEN]), llDetail);

  try
    FBroker.CallServer(RPC_GET_MESSAGE_DETAIL, [IEN], nil);
    result := FBroker.Results.Text;
    Log(Format('Message %s retrieved', [IEN]), llDetail);
  except
    on E: EBrokerError do
      begin
        result := '';
        Log(Format('Broker error %s received during broker call %s.', [E.Message, RPC_GET_MESSAGE_DETAIL]), llCritical);
      end;
  end;
end;

procedure TCPGatewaySvc.Log(const AMessage: string; const ALogLevel: TLogLevel);
begin
  {Always log llImportant}
  if (ord(ALogLevel) <= FLogLevel) or (aLogLevel = llImportant) then
    case aLogLevel of
      llCritical: LogMessage(aMessage, EVENTLOG_ERROR_TYPE);
      llError: LogMessage(aMessage, EVENTLOG_ERROR_TYPE);
      llWarning: LogMessage(aMessage, EVENTLOG_WARNING_TYPE);
      llImportant: LogMessage(aMessage, EVENTLOG_INFORMATION_TYPE);
      llDetail: LogMessage(aMessage, EVENTLOG_INFORMATION_TYPE);
    end;
end;

function TCPGatewaySvc.GenHL7_XMLMessage(AHL7Message: TCPHL7Message): ICliOFilerDoc;
var
  HL7Message: THL7Message;
  ObsSet: ICliOFilerSet;
  Obs: ICliOFilerObs;
  ObsDateTimeVal: string;
  ObsDateTime: TDateTime;
  I: Integer;
  TermMapper: ITermMapper;
  ObservationGroup: TStringList;
  CompleteMessage: TStringList;
  LocationID: string;
  TermID: string;
  SecondaryDelim: char;
  UoM: string;
  SubChar: char;
  ObsID: string;
  DebugInfo: string;
  //
  // For reporting porpoises.
  Method: string;
  Position: string;
  Location: string;
  Quality: string;
  Product: string;
  NullFlavor: string;
  Task: string;
  Schedule: string;
  Value: string;
begin
  FMappingErrorList.Clear; {All mapping errors get dumped here}
  Log('Generating HL7 Message', llDetail);

  HL7Message := THL7Message.Create(GetMessage(AHL7Message.ID), self as ICPGatewayLogger);

  try
    SubChar := HL7Message.EncodingChars[1];

    {HL7 Sending Application should match the term mapping table SOURCE_ID}
    TermMapper := FTerminologySource.GetTermMapperHL7(HL7Message.SendingApplication);

    if Assigned(TermMapper) then
      Log(Format('Using Term Mapper %s with ID: %s', [TermMapper.getName, TermMapper.getID]), llDetail)
    else
      raise Exception.CreateFmt('No term mapper returned for ''%s''.', [HL7Message.SendingApplication]);

    FAlarmList.Load(HL7Message.DFN);
    Result := FObservationFiler.getNewDocument(TermMapper);

    Log('Parsing OBX segments.', llDetail);

    CompleteMessage := TStringList.Create;
    try
      CompleteMessage.Text := HL7Message.Text;

      if Assigned(HL7Message.Segment['PV1']) then
        begin
          LocationID := Piece(HL7Message.Segment['PV1'].Text, HL7Message.Delimiter, 3);
          SecondaryDelim := HL7Message.EncodingChars[1];
          LocationID := Piece(LocationID, SecondaryDelim, 1);
        end
      else
        begin
          LocationID := '';
        end;

      while CompleteMessage.Count > 0 do
        begin
          ObservationGroup := GetObservationGroup(CompleteMessage);
          try
            if ObservationGroup.Count > 0 then
              begin
                Log('Starting on an Observation Group', llDetail);
                // First, we deal with the OBR stuff

                //Now, we'll want to generate a set with what was in OBR-4.
                ObsSet := Result.CreateNewSet;
                ObsSet.setEnteredDateTime(now);
                ObsSet.setComment(Piece(Piece(ObservationGroup[0], HL7Message.Delimiter, 5), '^', 1));

                for I := 1 to ObservationGroup.Count - 1 do
                  begin
                    if piece(ObservationGroup[I], HL7Message.Delimiter, 1) = OBR then
                      Continue;

                    //Note: For right now, we're not going to store ??? for unobserved
                    // observations.  This may (most likely will) change at some point.
                    if piece(ObservationGroup[I], HL7Message.Delimiter, 6) = '???' then
                      Continue;

                    if piece(ObservationGroup[I], HL7Message.Delimiter, 6) = '---' then
                      Continue;

                    if piece(ObservationGroup[I], HL7Message.Delimiter, 6) = '+++' then
                      Continue;

                    {
                    We have to be able to find the term itself.  If we can't, we are hosed.
                    }
                    ObsID := piece(ObservationGroup[I], HL7Message.Delimiter, 4);
                    ObsID := piece(ObsID, SubChar);
                    TermID := TermMapper.getTerm(ttObservation, ObsID);
                    if TermID = DO_NOT_PROCESS then
                      Continue;

                    //ttUnit, ttMethod, ttPosition, ttLocation, ttQuality, ttProduct, ttNullFlavor, ttTask, ttSchedule, ttValue
                    Obs := Result.CreateNewObs(ObsSet.getID);
                    Obs.AddToSet(ObsSet.getID);

                    Obs.setFacilityID(HL7Message.FacilityID);
                    Obs.setFacilityLocationID(LocationID);
                    Obs.setPatientID(HL7Message.DFN);
                    Obs.setEnteredDateTime(Now);


                    Obs.setSource(HL7Message.SendingApplication);
                    DebugInfo :=
                      Piece(ObservationGroup[I], HL7Message.Delimiter, 4) + '|' +
                      Piece(ObservationGroup[I], HL7Message.Delimiter, 5) + '|' +
                      Piece(ObservationGroup[I], HL7Message.Delimiter, 6) + '|' +
                      Piece(ObservationGroup[I], HL7Message.Delimiter, 7);
                    DebugInfo := StringReplace(DebugInfo, '^', '\u094', [rfReplaceAll]);
                    DebugInfo := Copy(DebugInfo, 1, 50);
                    Obs.setSourceDataItemID(DebugInfo);
                    Obs.setSourceVersion(HL7Message.Version);
                    Obs.setSourceComment('CPGatewayService.exe');
                    Log(format('New Observation ID: %s', [Obs.getID]), llDetail);


                    ObsDateTimeVal := Piece(ObservationGroup[I], HL7Message.Delimiter, 15);
                    if ObsDateTimeVal <> '' then
                      ObsDateTime := HL7DateTimeToTDateTime(ObsDateTimeVal)
                    else
                      ObsDateTime := now;

                    Obs.setObservedDateTime(ObsDateTime);
                    Obs.setValue(Piece(ObservationGroup[I], HL7Message.Delimiter, 6));

                    Obs.setTermID(TermID);

                    Log(format('Incoming term %s mapped to %s.', [
                     ObsID, Obs.getTermID]), llDetail);

                    UoM := piece(ObservationGroup[I], HL7Message.Delimiter, 7);
                    UoM := piece(UOM, SubChar);

                    if Length(UoM) > 0 then {If unit is in HL7 seg 7 it *MUST* be used}
                      try
                        Obs.setQualifier(ttUnit, TermMapper.getTerm(ttUnit, UoM));
                        Log(format('HL7 unit %s mapped to %s.', [UoM, Obs.getQualifier(ttUnit)]), llDetail);
                      except on E: ETerminologyMapperError do
                          begin
                            FMappingErrorList.Add(Format('HL7 unit ''%s'' for Vendor Key ''%s'' has no mapping.', [UoM, ObsID]));
                            Log(Format('HL7 unit ''%s'' for Vendor Key ''%s'' has no mapping.', [UoM, ObsID]), llWarning);
                          end
                      end
                    else {There was no actual unit passed in, use the obs code and try that angle}
                      try
                        UoM := piece(ObservationGroup[I], HL7Message.Delimiter, 4);
                        UoM := TermMapper.getTerm(ttUnit, UoM);
                        Obs.setQualifier(ttUnit, TermMapper.getTerm(ttUnit, UoM));
                        Log(format('Incoming unit %s mapped to %s.', [UoM, Obs.getQualifier(ttUnit)]), llDetail);
                      except
                        Obs.setQualifier(ttUnit, '');
                      end;

                    try
                      Method := TermMapper.getTerm(ttMethod, ObsID);
                      Obs.setQualifier(ttMethod, Method);
                      Log(format('Incoming method %s mapped to %s.', [ObsID, Method]), llDetail);
                    except on E: ETerminologyMapperError do
                    end;

                    try
                      Position := TermMapper.getTerm(ttPosition, ObsID);
                      Obs.setQualifier(ttPosition, Position);
                      Log(format('Incoming position %s mapped to %s.', [ObsID, Position]), llDetail);
                    except on E: ETerminologyMapperError do
                    end;

                    try
                      Location := TermMapper.getTerm(ttLocation, ObsID);
                      Obs.setQualifier(ttLocation, Location);
                      Log(format('Incoming location %s mapped to %s.', [ObsID, Location]), llDetail);
                    except on E: ETerminologyMapperError do
                    end;

                    try
                      Quality := TermMapper.getTerm(ttQuality, ObsID);
                      Obs.setQualifier(ttQuality, Quality);
                      Log(format('Incoming quality %s mapped to %s.', [ObsID, Quality]), llDetail);
                    except on E: ETerminologyMapperError do
                    end;

                    try
                      Product := TermMapper.getTerm(ttProduct, ObsID);
                      Obs.setQualifier(ttProduct, Product);
                      Log(format('Incoming product %s mapped to %s.', [ObsID, Product]), llDetail);
                    except on E: ETerminologyMapperError do
                    end;

                    try
                      NullFlavor := TermMapper.getTerm(ttNullFlavor, ObsID);
                      Obs.setQualifier(ttNullFlavor, NullFlavor);
                      Log(format('Incoming Null Flavor %s mapped to %s.', [ObsID, NullFlavor]), llDetail);
                    except on E: ETerminologyMapperError do
                    end;

                    try
                      Quality := TermMapper.getTerm(ttQuality, ObsID);
                      Obs.setQualifier(ttQuality, Quality);
                      Log(format('Incoming Quality %s mapped to %s.', [ObsID, Quality]), llDetail);
                    except on E: ETerminologyMapperError do
                    end;

                    try
                      Schedule := TermMapper.getTerm(ttSchedule, ObsID);
                      Obs.setQualifier(ttSchedule, Schedule);
                      Log(format('Incoming Schedule %s mapped to %s.', [ObsID, Schedule]), llDetail);
                    except on E: ETerminologyMapperError do
                    end;

                    try
                      Value := TermMapper.getTerm(ttValue, ObsID);
                      Obs.setQualifier(ttValue, Value);
                      Log(format('Incoming Value %s mapped to %s.', [ObsID, Value]), llDetail);
                    except on E: ETerminologyMapperError do
                    end;

                  end;
              end;
          finally
            FreeAndNil(ObservationGroup);
          end;
        end;
    finally
      FreeAndNil(CompleteMessage);
    end;
  finally
    FreeAndNil(HL7Message);
  end;
end;

procedure TCPGatewaySvc.ChangeMessageStatus(AMessage: TCPHL7Message; const Status: string; const StatusReason: TStringList = nil);
var
  I: Integer;
begin
  try
    FBroker.CallServer(RPC_UPDATE_STATUS, [AMessage.ID, Status], nil);
    if Assigned(StatusReason) then
      begin
        // Note: Dan's handing me back some prettyprint stuff to make the error
        // messages more readable for end users.  We should probably get rid of it.
        //
        // We may need to fudge factor the lead-in here: According to my count, I should
        // be recieving 5 lines before the meat of the message if it is a validation
        // error.  So, anything else needs to get padded up so that there are at least
        // six lines there.
        for I := 0 to StatusReason.Count - 1 do
          begin
            if MatchesMask(StatusReason[I], 'Error validating Filer Document.') then
              continue;
            if MatchesMask(StatusReason[I], 'CliO Filer Error Log') then
              continue;
            if MatchesMask(StatusReason[I], '*error(s) found in log.') then
              continue;
            if Trim(StatusReason[I]) = '' then
              continue;
            FBroker.CallServer(RPC_UPDATE_REASON, [AMessage.ID, StatusReason[I]], nil);
          end;
      end;
  except on E: Exception do
      begin
        Log(Format('Unable to change message status for message %s: %s', [AMessage.ID, E.Message]) + '[' + e.ClassName + ']', llError);
      end;
  end;
end;

procedure TCPGatewaySvc.Cleanup;
begin
  fCPParameters := nil;
  FConfiguration := nil;
  FDBConnection := nil;
  FTerminologySource := nil;
  FCliOService := nil;
  FObservationFiler := nil;
  FBroker := nil;

  if Assigned(fNotifier) then
    FreeAndNil(FNotifier);
  if Assigned(fMappingErrorList) then
    FreeAndNil(FMappingErrorList);
end;

procedure TCPGatewaySvc.acPollForMessagesExecute(Sender: TObject);
begin
  FNotifier.Active := true;
end;

end.

