unit oTermUnitPair;

interface

uses
  ComCtrls,
  XMLIntf,
  SysUtils,
  Classes,
  uXML,
  oTermComponent,
  iCPTerminology;

type
  TTermUnitPair = class(TTermComponent, ITermComponent, ITermUnitPair)
  private
    fDecPrecision: integer;
    fMaxValue: double;
    fMinValue: double;
    fReferenceHigh: double;
    fReferenceLow: double;
  public
    constructor Create(aParentTerm: ITerm; aUnit: ITerm);
    destructor Destroy; override;

    function getDecPrecision: integer; stdcall;
    function getMaxValue: double; stdcall;
    function getMinValue: double; stdcall;
    function getReferenceHigh: double; stdcall;
    function getReferenceLow: double; stdcall;
    function getComponentDisplayName: string; override; stdcall;

    procedure setDecPrecision(const aValue: integer); stdcall;
    procedure setMaxValue(const aValue: double); stdcall;
    procedure setMinValue(const aValue: double); stdcall;
    procedure setReferenceHigh(const aValue: double); stdcall;
    procedure setReferenceLow(const aValue: double); stdcall;

    procedure Load(aXMLNode: IXMLNode); override; stdcall;
    procedure Save(aXMLNode: IXMLNode); override; stdcall;
    procedure Validate; override; stdcall;
  end;

implementation

const
  {Default XML Tags for the properties}
  XML_MIN_VALUE = 'MIN_VALUE';
  XML_MAX_VALUE = 'MAX_VALUE';
  XML_DEC_PRECISION = 'DEC_PRECISION';
  XML_REFERENCE_LOW = 'REFERENCE_LOW';
  XML_REFERENCE_HIGH = 'REFERENCE_HIGH';

  {Default values for the properties}
  DEFAULT_MIN_VALUE = 0.0;
  DEFAULT_MAX_VALUE = 100.0;
  DEFAULT_DEC_PRECISION = 0;
  DEFAULT_REFERENCE_LOW = 20.0;
  DEFAULT_REFERENCE_HIGH = 80.0;

constructor TTermUnitPair.Create(aParentTerm: ITerm; aUnit: ITerm);
begin
  inherited Create(ctUnit, aParentTerm, aUnit);
  fDecPrecision := DEFAULT_DEC_PRECISION;
  fMinValue := DEFAULT_MIN_VALUE;
  fReferenceLow := DEFAULT_REFERENCE_LOW;
  fReferenceHigh := DEFAULT_REFERENCE_HIGH;
  fMaxValue := DEFAULT_MAX_VALUE;
end;

destructor TTermUnitPair.Destroy;
begin
  inherited;
end;

procedure TTermUnitPair.Load(aXMLNode: IXMLNode);
begin
  try
    inherited Load(aXMLNode);
    fDecPrecision := XMLValueAsInt(aXMLNode, XML_DEC_PRECISION, DEFAULT_DEC_PRECISION);
    fMinValue := XMLValueAsDouble(aXMLNode, XML_MIN_VALUE, DEFAULT_MIN_VALUE);
    fReferenceLow := XMLValueAsDouble(aXMLNode, XML_REFERENCE_LOW, DEFAULT_REFERENCE_LOW);
    fReferenceHigh := XMLValueAsDouble(aXMLNode, XML_REFERENCE_HIGH, DEFAULT_REFERENCE_HIGH);
    fMaxValue := XMLValueAsDouble(aXMLNode, XML_MAX_VALUE, DEFAULT_MAX_VALUE);
  except
    on E: Exception do
      raise ETerminologyLoadError.CreateFmt('Unable to load %s. Error: %s', [Self.ClassName, E.Message]);
  end;
end;

procedure TTermUnitPair.Save(aXMLNode: IXMLNode);
begin
  try
    inherited Save(aXMLNode);
    aXMLNode.ChildValues[XML_DEC_PRECISION] := fDecPrecision;
    aXMLNode.ChildValues[XML_MIN_VALUE] := fMinValue;
    aXMLNode.ChildValues[XML_REFERENCE_LOW] := fReferenceLow;
    aXMLNode.ChildValues[XML_REFERENCE_HIGH] := fReferenceHigh;
    aXMLNode.ChildValues[XML_MAX_VALUE] := fMaxValue;
  except
    on E: Exception do
      raise ETerminologyLoadError.CreateFmt('Unable to save %s. Error: %s', [Self.ClassName, E.Message]);
  end;
end;

{ Private Getters }

function TTermUnitPair.getComponentDisplayName: string;
begin
  Result := getComponentTerm.getTerm;
end;

function TTermUnitPair.getDecPrecision: integer;
begin
  Result := fDecPrecision;
end;

function TTermUnitPair.getMaxValue: double;
begin
  Result := fMaxValue;
end;

function TTermUnitPair.getMinValue: double;
begin
  Result := fMinValue;
end;

function TTermUnitPair.getReferenceHigh: double;
begin
  Result := fReferenceHigh;
end;

function TTermUnitPair.getReferenceLow: double;
begin
  Result := fReferenceLow;
end;

{ Private Setters }

procedure TTermUnitPair.setDecPrecision(const aValue: integer);
begin
  fDecPrecision := aValue;
end;

procedure TTermUnitPair.setMaxValue(const aValue: double);
begin
  fMaxValue := aValue;
end;

procedure TTermUnitPair.setMinValue(const aValue: double);
begin
  fMinValue := aValue;
end;

procedure TTermUnitPair.setReferenceHigh(const aValue: double);
begin
  fReferenceHigh := aValue;
end;

procedure TTermUnitPair.setReferenceLow(const aValue: double);
begin
  fReferenceLow := aValue;
end;

procedure TTermUnitPair.Validate;
begin
  inherited;
  if getComponentTerm = nil then
    begin
      getParentTerm.NewError(etComponentSetting, 'No unit assigned to this Unit-Pair');
      Exit;
    end;

  if getComponentTerm.getTermType <> ttUnit then
    getParentTerm.NewError(etComponentSetting, ' non unit term assigned to this Unit-Pair');

  if (fDecPrecision < 0) or (fDecPrecision > 8) then
    getParentTerm.NewError(etComponentSetting, getComponentTerm.getTerm + ' decimal precision setting not 0-8');

  if fMinValue >= fReferenceLow then
    getParentTerm.NewError(etComponentSetting, getComponentTerm.getTerm + ' minimum value must be less than reference low value');

  if fReferenceLow >= fReferenceHigh then
    getParentTerm.NewError(etComponentSetting, getComponentTerm.getTerm + ' reference low value must be less than reference high value');

  if fReferenceHigh >= fMaxValue then
    getParentTerm.NewError(etComponentSetting, getComponentTerm.getTerm + ' reference high value must be less than maximum value');
end;

end.

