unit oTerminologyController;

interface

uses
  uXML,
  ComCtrls,
  XMLIntf,
  XMLDoc,
  SysUtils,
  Classes,
  iCPTerminology;

type
  TTerminologyController = class(TInterfacedObject, ITerminologyController)
  private
    fFileName: string;

    fTermNode: IXMLNode;
    fTermList: IInterfaceList;

    fXMLDoc: IXMLDocument;
    fHeaderNode: IXMLNode;

    fOnStatusUpdate: TOnStatusUpdate;

    function getDescription: string; stdcall;
    function getDateTimeCreated: TDateTime; stdcall;
    function getDateTimeModified: TDateTime; stdcall;

    function IsOpen: boolean; stdcall;
    function getCount: integer; stdcall;
    function getCountByType(aTermType: TTermType): integer; stdcall;
    function getTerm(aID: string): ITerm; stdcall;
    function getTermName(aID: string): string; stdcall;
    function getEnumerator: ITerminologyEnumerator; stdcall;
    function getModified: boolean; stdcall;
    function getFileName: string; stdcall;
    function getNewTerm(aTermType: TTermType; aDataType: TDataType = dtNone): ITerm; stdcall;

    procedure setDescription(const aValue: string); stdcall;

    procedure LoadList(aList: TStrings); overload; stdcall;
    procedure LoadList(aList: TStrings; aTermType: TTermType); overload; stdcall;

    function LoadTermAttributes(aList: TStrings): integer; stdcall;
    function LoadTermTypes(aList: TStrings): integer; stdcall;
    function LoadDataTypes(aList: TStrings): integer; stdcall;
    function LoadToDoStatuses(aList: TStrings): integer; stdcall;

    procedure Load(aFileName: string); stdcall;
    procedure Close; stdcall;

    procedure Save; stdcall;
    procedure SaveAs(aFileName: string); stdcall;

    procedure setOnStatusUpdate(aProcedure: TOnStatusUpdate); stdcall;

  public
    constructor Create;
    destructor Destroy; override;
  end;

implementation

uses
  oTerm,
  oTerminologyEnumerator;

{ TTermList }

constructor TTerminologyController.Create;
begin
  inherited Create;
  TInterfaceList.Create.GetInterface(IInterfaceList, fTermList);
end;

destructor TTerminologyController.Destroy;
begin
  fTermList.Clear;
  fTermList := nil;
  inherited;
end;

function TTerminologyController.getCount: integer;
begin
  Result := fTermList.Count;
end;

function TTerminologyController.getCountByType(aTermType: TTermType): integer;
begin
  Result := 0;
  with getEnumerator do
    while MoveNext do
      if GetCurrentTerm.getTermType = aTermType then
        inc(Result);
end;

function TTerminologyController.getDateTimeCreated: TDateTime;
begin
  Result := XMLValueAsDateTime(fHeaderNode, 'DATE_TIME_CREATED');
end;

function TTerminologyController.getDateTimeModified: TDateTime;
begin
  Result := XMLValueAsDateTime(fHeaderNode, 'DATE_TIME_MODIFIED');
end;

function TTerminologyController.getDescription: string;
begin
  Result := XMLValueAsStr(fHeaderNode, 'DESCRIPTION');
end;

function TTerminologyController.getModified: boolean;
begin
  if Assigned(fXMLDoc) then
    Result := fXMLDoc.Modified
  else
    Result := False;
end;

function TTerminologyController.getNewTerm(aTermType: TTermType; aDataType: TDataType = dtNone): ITerm;
begin
  if TTerm.Create(self as ITerminologyController, fTermNode.AddChild('RECORD')).GetInterface(ITerm, Result) then
    with Result do
      begin
        setTerm('NEW TERM');
        setAbbreviation('nt');
        setDisplayName('New Term');
        setDescription('No Description');
        setHelpText('No Help Text');
        setTermType(aTermType);
        setDataType(aDataType);
        setActive(False);
        fTermList.Add(Result);
      end
  else
    Result := nil;
end;

function TTerminologyController.getTerm(aID: string): ITerm;
begin
  if aID = '' then
    begin
      Result := nil;
      Exit;
    end
  else
    with getEnumerator do
      while MoveNext do
        if GetCurrentTerm.getID = aID then
          begin
            getCurrentTerm.QueryInterface(ITerm, Result);
            Exit;
          end;

  Result := nil;
  raise ETerminologySaveError.CreateFmt('Unable to locate term ID: %s', [aID]);
end;

function TTerminologyController.getEnumerator: ITerminologyEnumerator;
begin
  TTerminologyEnumerator.Create(fTermList).GetInterface(ITerminologyEnumerator, Result);
end;

function TTerminologyController.getTermName(aID: string): string;
var
  aTerm: ITerm;
begin
  aTerm := getTerm(aID);
  if Assigned(aTerm) then
    Result := aTerm.getTerm
  else
    Result := '';
end;

procedure TTerminologyController.Load(aFileName: string);
var
  i, j: integer;
  aNode: IXMLNode;
  aTerm: ITerm;
  aTerm1, aTerm2: ITerm; {Sorters}
begin
  try
    fTermList.Clear;
    fXMLDoc := NewXMLDocument;
    fFileName := aFileName;

    if not FileExists(fFileName) then
      try
        fXMLDoc.DocumentElement := fXMLDoc.AddChild('TERMINOLOGY');
        with fXMLDoc.DocumentElement.AddChild('HEADER') do
          begin
            ChildValues['DESCRIPTION'] := 'Clinical Procedures Terminology File';
            ChildValues['DATE_TIME_CREATED'] := DateTimeToXMLString(Now);
            ChildValues['DATE_TIME_MODIFIED'] := DateTimeToXMLString(Now);
          end;
        fXMLDoc.DocumentElement.AddChild('TERMS');
        fXMLDoc.SaveToFile(fFileName);
      except
        raise Exception.Create('Unable to create new terminology file.');
      end
    else
      try
        fXMLDoc.LoadFromFile(aFileName);
      except
        raise Exception.Create('Error loading terminology file.');
      end;

    fXMLDoc.Active := True;
    fTermNode := fXMLDoc.DocumentElement.ChildNodes.FindNode('TERMS');
    fHeaderNode := fXMLDoc.DocumentElement.ChildNodes.FindNode('HEADER');

    {Once through to build the dictionary}
    aNode := fTermNode.ChildNodes.First;
    i := 0;
    while Assigned(aNode) do
      try
        inc(i);
        if Assigned(fOnStatusUpdate) then
          fOnStatusUpdate(Format('Loading terminology dictionary - %s Complete...', [getPercentageDone(i, fTermNode.ChildNodes.Count)]));
        if TTerm.Create(Self as ITerminologyController, aNode).GetInterface(ITerm, aTerm) then
          fTermList.Add(aTerm)
        else
          raise ETerminologyDictionaryError.Create('Unable to create term' + #13 + aNode.XML);
      finally
        aNode := aNode.NextSibling;
      end;

    {Now go through so that each term can build it's own components}
    with getEnumerator do
      while MoveNext do
        begin
          if Assigned(fOnStatusUpdate) then
            fOnStatusUpdate(Format('Loading Clinical Data Model - %d%% Complete...', [getPercentageDone]));
          GetCurrentTerm.Load;
        end;

    { Old School Sort by Component Order :) }
    if fTermList.Count > 1 then
      begin
        if Assigned(fOnStatusUpdate) then
          fOnStatusUpdate('Sorting Dictionary...');

        for i := 0 to fTermList.Count - 2 do
          for j := i to fTermList.Count - 1 do
            begin
              fTermList[i].QueryInterface(ITerm, aTerm1);
              fTermList[j].QueryInterface(ITerm, aTerm2);
              if aTerm1.getTerm > aTerm2.getTerm then
                fTermList.Exchange(i, j);
            end;
      end;

    if Assigned(fOnStatusUpdate) then
      fOnStatusUpdate;
  except
    on E: Exception do
      raise Exception.CreateFmt('Error loading file %s. Error: %s', [aFileName, E.Message]);
  end;
end;

procedure TTerminologyController.LoadList(aList: TStrings; aTermType: TTermType);
begin
  aList.Clear;
  with getEnumerator do
    while MoveNext do
      if GetCurrentTerm.getTermType = aTermType then
        GetCurrentTerm.AddToList(aList);
end;

procedure TTerminologyController.LoadList(aList: TStrings);
begin
  aList.Clear;
  with GetEnumerator do
    while MoveNext do
      GetCurrentTerm.AddToList(aList);
end;

procedure TTerminologyController.Save;
begin
  fHeaderNode.ChildValues['DATE_TIME_MODIFIED'] := DateTimeToXMLString(Now);
  fXMLDoc.SaveToFile(fFileName);
end;

procedure TTerminologyController.SaveAs(aFileName: string);
begin
  fHeaderNode.ChildValues['DATE_TIME_CREATED'] := DateTimeToXMLString(Now);
  fHeaderNode.ChildValues['DATE_TIME_MODIFIED'] := DateTimeToXMLString(Now);
  fFileName := aFileName;
  Save;
end;

procedure TTerminologyController.setDescription(const aValue: string);
begin
  fHeaderNode.ChildValues['DESCRIPTION'] := aValue;
end;

procedure TTerminologyController.setOnStatusUpdate(aProcedure: TOnStatusUpdate);
begin
  fOnStatusUpdate := aProcedure;
end;

procedure TTerminologyController.Close;
begin
  try
    fTermList.Clear;
    fFileName := '';

    fXMLDoc := nil;
    fHeaderNode := nil;
    fTermNode := nil;
  except
    on E: Exception do
      raise Exception.CreateFmt('Error closing file. Error: %s', [E.Message]);
  end;
end;

function TTerminologyController.IsOpen: boolean;
begin
  Result := (fFileName <> '');
end;

function TTerminologyController.getFileName: string;
begin
  Result := fFileName;
end;

function TTerminologyController.LoadTermAttributes(aList: TStrings): integer;
var
  aAttr: TTermAttribute;
begin
  aList.Clear;
  for aAttr := Low(TTermAttribute) to High(TTermAttribute) do
    aList.Add(TERM_ATTRIBUTE[aAttr]);
  Result := aList.Count;
end;

function TTerminologyController.LoadTermTypes(aList: TStrings): integer;
var
  aType: TTermType;
begin
  aList.Clear;
  for aType := Low(TTermType) to High(TTermType) do
    aList.Add(TERM_TYPE[aType]);
  Result := aList.Count;
end;

function TTerminologyController.LoadDataTypes(aList: TStrings): integer;
var
  aType: TDataType;
begin
  aList.Clear;
  for aType := Low(TDataType) to High(TDataType) do
    aList.Add(TERM_DATATYPE[aType]);
  Result := aList.Count;
end;

function TTerminologyController.LoadToDoStatuses(aList: TStrings): integer;
var
  aType: TTermToDoStatus;
begin
  aList.Clear;
  for aType := Low(TTermToDoStatus) to High(TTermToDoStatus) do
    aList.Add(TERM_TODO_STATUS[aType]);
  Result := aList.Count;
end;

end.

