unit uScenario;

interface

uses
  Classes, SysUtils, ExtCtrls, uPatients, uDevices;

type
  TTimedObservationItem = class;
  TTimedObservationEvent = procedure(Sender: TTimedObservationItem) of object;
  
  TTimedObservationItem = class(TCollectionItem)
  private
    FPatient: TPatient;
    FObservation: TObservation;
    FDevice: TDevice;
    FTimer: TTimer;
    FOnObservation: TTimedObservationEvent;
  protected
    procedure SetInterval(const Value: Integer);
    function GetInterval: Integer;
    procedure DoTimer(Sender: TObject);
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
  published
    property Patient: TPatient read FPatient write FPatient;
    property Observation: TObservation read FObservation write FObservation;
    property Device: TDevice read FDevice write FDevice;
    property Interval: Integer read GetInterval write SetInterval;
    property OnObservation: TTimedObservationEvent read FOnObservation write FOnObservation;
  end;

  TTimedObservationList = class(TCollection)
  private
  protected
    function GetItem(Index: Integer): TTimedObservationItem;
    procedure SetItem(Index: Integer; Value: TTimedObservationItem);
  public
    function Add: TTimedObservationItem;
    procedure Assign(Source: TPersistent); override;
    function FindItemID(ID: Integer): TTimedObservationItem;
    function Insert(Index: Integer):TTimedObservationItem;
    property Items[Index: Integer]: TTimedObservationItem read GetItem write SetItem; default;
  published
  end;

  TBed = class(TCollectionItem)
  private
    FOnObservation: TTimedObservationEvent;
    FTimedObservations: TTimedObservationList;
    FOnChange: TNotifyEvent;
    FPatient: TPatient;
    FDevices: TUnownedDeviceList;
  protected
    procedure DoObservation(Sender: TTimedObservationItem);
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure Start;
    procedure Stop;
  published
    property OnObservation: TTimedObservationEvent read FOnObservation write FOnObservation;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property Patient: TPatient read FPatient write FPatient;
    property Devices: TUnownedDeviceList read FDevices write FDevices;
  end;

  TWard = class(TCollection)
  private
    FOnChange: TNotifyEvent;
    FName: string;
  protected
    function GetItem(Index: Integer): TBed;
    procedure SetItem(Index: Integer; Value: TBed);
  public
    function Add: TBed;
    procedure Assign(Source: TPersistent); override;
    function FindItemID(ID: Integer): TBed;
    function FindItemByPatient(const APatient: TPatient): TBed;
    function Insert(Index: Integer):TBed;
    property Items[Index: Integer]: TBed read GetItem write SetItem; default;
  published
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property Name: string read FName write FName;
  end;


  TScenario = class(TCollectionItem)
  private
    FWard: TWard;
  protected
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure Start;
    procedure Stop;
  published
    property Ward: TWard read FWard write FWard;
  end;

  TScenarioList = class(TCollection)
  private
    function GetItem(Index: Integer): TScenario;
    procedure SetItem(Index: Integer; Value: TScenario);
  public
    function Add: TScenario;
    procedure Assign(Source: TPersistent); override;
    function FindItemID(ID: Integer): TScenario;
    function FindItemName(const AName: string): TScenario;
    function Insert(Index: Integer):TScenario;
    procedure Save(const AFilename: string);
    procedure Load(const AFilename: string; const APatients: TPatientList; const ADevices: TDeviceList);
    property Items[Index: Integer]: TScenario read GetItem write SetItem; default;
  published
  end;

implementation

uses
  XMLDoc, XMLIntf, XMLDom, Variants, uConstants;

{ TWard }

function TWard.Add: TBed;
begin
  result := TBed(inherited Add);
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TWard.Assign(Source: TPersistent);
var
  I: Integer;
  SourceWard: TWard;
  Bed: TBed;
begin
  if Source is TWard then
  begin
    Clear;
    SourceWard := TWard(Source);
    for I := 0 to SourceWard.Count - 1 do
    begin
      Bed := Add;
      Bed.Assign(SourceWard[I]);
    end;
  end
  else
    inherited;
end;

function TWard.FindItemByPatient(const APatient: TPatient): TBed;
var
  I: Integer;
begin
  result := nil;
  for I := 0 to Count - 1 do
    if Items[I].Patient.ID = APatient.ID then
    begin
      result := Items[I];
      Break;
    end;
end;

function TWard.FindItemID(ID: Integer): TBed;
begin
  result := TBed(inherited FindItemID(ID));
end;

function TWard.GetItem(Index: Integer): TBed;
begin
  result := TBed(inherited GetItem(Index));
end;

function TWard.Insert(Index: Integer): TBed;
begin
  result := TBed(inherited Insert(Index));
  if Assigned(FOnChange) then
    FOnChange(Self);

end;

procedure TWard.SetItem(Index: Integer; Value: TBed);
begin
  inherited SetItem(Index, Value);
end;

{ TBed }

procedure TBed.Assign(Source: TPersistent);
var
  SourceBed: TBed;
begin
  if Source is TBed then
  begin
    FPatient := nil;
    FDevices.Clear;
    
    SourceBed := TBed(Source);
    FOnObservation := SourceBed.FOnObservation;
    FPatient := SourceBed.Patient;
    FDevices.Assign(SourceBed.FDevices);
  end
  else
    inherited;
end;

constructor TBed.Create(Collection: TCollection);
begin
  inherited;
  FDevices := TUnownedDeviceList.Create;
  FTimedObservations := TTimedObservationList.Create(TTimedObservationItem);
end;

destructor TBed.Destroy;
begin
  FreeAndNil(FTimedObservations);
  FreeAndNil(FDevices);
  inherited;
end;

procedure TBed.DoObservation(Sender: TTimedObservationItem);
begin
  if Assigned(FOnObservation) then
  begin
    FOnObservation(Sender);
  end;
end;

procedure TBed.Start;
var
  ObsEnum : TCollectionEnumerator;
begin
  ObsEnum := FTimedObservations.GetEnumerator;
  while ObsEnum.MoveNext do
    TTimedObservationItem(ObsEnum.Current).FTimer.Enabled := true;
end;

procedure TBed.Stop;
var
  ObsEnum : TCollectionEnumerator;
begin
  ObsEnum := FTimedObservations.GetEnumerator;
  while ObsEnum.MoveNext do
    TTimedObservationItem(ObsEnum.Current).FTimer.Enabled := false;
end;

{ TScenarioList }

function TScenarioList.Add: TScenario;
begin
  result := TScenario(inherited Add);
end;

procedure TScenarioList.Assign(Source: TPersistent);
var
  I: Integer;
  SList: TScenarioList;
  Scenario: TScenario;
begin
  if Source is TScenarioList then
  begin
    SList := TScenarioList(Source);
    Clear;

    for I := 0 to SList.Count - 1 do
    begin
      Scenario := Add;
      Scenario.Assign(SList[I]);
    end;
  end
  else
    inherited;
end;

function TScenarioList.FindItemID(ID: Integer): TScenario;
begin
  result := TScenario(inherited FindItemID(ID));
end;

function TScenarioList.FindItemName(const AName: string): TScenario;
var
  I: Integer;
begin
  result := nil;
  for I := 0 to Count - 1 do
    if Items[I].Ward.Name = AName then
    begin
      result := Items[I];
      Break;
    end;
end;

function TScenarioList.GetItem(Index: Integer): TScenario;
begin
  result := TScenario(inherited GetItem(Index));
end;

function TScenarioList.Insert(Index: Integer): TScenario;
begin
  result := TScenario(inherited Insert(Index));
end;

procedure TScenarioList.Load(const AFilename: string; const APatients: TPatientList; const ADevices: TDeviceList);
var
  XMLDoc: IXMLDocument;
  RootNode: IXMLNode;
  ChildNode: IXMLNode;
  BedsNode: IXMLNode;
  BedNode: IXMLNode;
  WardNode: IXMLNode;
  I, J, K, L: Integer;
  AScenario: TScenario;
  ABed: TBed;
  ADevice: TDevice;
  ObsItem: TTimedObservationItem;
begin
  Clear;

  XMLDoc := TXMLDocument.Create(nil);
  try
    DefaultDOMVendor := 'MSXML';

    XMLDoc.LoadFromFile(AFilename);
    XMLDoc.Options := XMLDoc.Options + [doNodeAutoIndent];
    XMLDoc.Active := true;
    XMLDoc.Encoding := 'UTF-8';

    RootNode := XMLDoc.DocumentElement;

    if RootNode.NodeName <> SCENARIOS then
      raise Exception.Create('Incorrect file format for scenario file!');

    for I := 0 to RootNode.ChildNodes.Count - 1 do
    begin
      ChildNode := RootNode.ChildNodes[I];
      WardNode := ChildNode.ChildNodes.FindNode(WARD);
      if assigned(ChildNode) then
      begin
        AScenario := Add;

        ChildNode := WardNode.ChildNodes.FindNode(NAME);
        if assigned(ChildNode) and (ChildNode.NodeValue <> null) then
          AScenario.Ward.Name := ChildNode.NodeValue;

        BedsNode := WardNode.ChildNodes.FindNode(BEDS);
        if assigned(BedsNode) then
        begin
          for J := 0 to BedsNode.ChildNodes.Count - 1 do
          begin
            BedNode := BedsNode.ChildNodes[J];
            ABed := AScenario.Ward.Add;

            ChildNode := BedNode.ChildNodes.FindNode(PATIENT);
            if assigned(ChildNode) and (ChildNode.NodeValue <> null) then
              ABed.FPatient := APatients.Find(ChildNode.NodeValue);

            if not assigned(ABed.FPatient) then
            begin
              AScenario.FWard.Delete(ABed.Index);
              Continue;
            end;

            ChildNode := BedNode.ChildNodes.FindNode(DEVICES);
            for K := 0 to ChildNode.ChildNodes.Count - 1 do
            begin
              if ChildNode.ChildNodes[K].NodeName = DEVICE then
              begin
                ADevice := ADevices.Find(ChildNode.ChildNodes[K].NodeValue);
                if Assigned(ADevice) then
                begin
                  if ADevice.SendAllObservationsAtOnce then
                  begin
                    ObsItem := ABed.FTimedObservations.Add;
                    ObsItem.FPatient := ABed.Patient;
                    ObsItem.FDevice := ADevice;
                    ObsItem.FObservation := nil;
                    ObsItem.Interval := ADevice.Interval;
                    ObsItem.OnObservation := ABed.DoObservation;
                  end
                  else
                  begin
                    for L := 0 to ADevice.Observations.Count - 1 do
                    begin
                      ObsItem := ABed.FTimedObservations.Add;
                      ObsItem.FPatient := ABed.Patient;
                      ObsItem.FDevice := ADevice;
                      ObsItem.FObservation := ADevice.Observations[L];
                      ObsItem.Interval := ADevice.Observations[L].Frequency;
                      ObsItem.OnObservation := ABed.DoObservation;
                    end;
                  end;
                  ABed.FDevices.Add(ADevice);
                end;
              end;
            end;
          end;
        end;
      end;
    end;
  finally
    XMLDoc := nil;
  end;
end;

procedure TScenarioList.Save(const AFilename: string);
var
  XMLDoc: IXMLDocument;
  RootNode: IXMLNode;
  ScenarioNode: IXMLNode;
  WardNode: IXMLNode;
  BedsNode: IXMLNode;
  BedNode: IXMLNode;
  DevicesNode: IXMLNode;
  I, J, K: Integer;
begin
  XMLDoc := TXMLDocument.Create(nil);
  try
    DefaultDOMVendor := 'MSXML';

    XMLDoc.Options := XMLDoc.Options + [doNodeAutoIndent];
    XMLDoc.Active := true;
    XMLDoc.Encoding := 'UTF-8';

    RootNode := XMLDoc.CreateElement(SCENARIOS, '');
    XMLDoc.DocumentElement := RootNode;

    for I := 0 to Count - 1 do
    begin
      ScenarioNode := RootNode.AddChild(SCENARIO);

      WardNode := ScenarioNode.AddChild(WARD);
      WardNode.AddChild(NAME).NodeValue := Items[I].FWard.FName;

      BedsNode := WardNode.AddChild(BEDS);
      for J := 0 to Items[I].FWard.Count - 1 do
      begin
        BedNode := BedsNode.AddChild(BED);
        BedNode.AddChild(PATIENT).NodeValue := Items[I].FWard.Items[J].FPatient.ID;

        DevicesNode := BedNode.AddChild(DEVICES);
        for K := 0 to Items[I].FWard.Items[J].FDevices.Count - 1 do
        begin
          DevicesNode.AddChild(DEVICE).NodeValue := Items[I].FWard.Items[J].FDevices[K].Description;
        end;
      end;
    end;
    XMLDoc.SaveToFile(AFilename);
  finally
    XMLDoc := nil;
  end;
end;

procedure TScenarioList.SetItem(Index: Integer; Value: TScenario);
begin
  inherited SetItem(Index, Value);
end;

{ TScenario }

procedure TScenario.Assign(Source: TPersistent);
begin
  if Source is TScenario then
  begin
    FWard.Assign(TScenario(Source).Ward);
  end
  else
    inherited;
end;

constructor TScenario.Create(Collection: TCollection);
begin
  inherited;
  FWard := TWard.Create(TBed);
end;

destructor TScenario.Destroy;
begin
  FreeAndNil(FWard);
  inherited;
end;

procedure TScenario.Start;
var
  I: Integer;
begin
{ DONE 1 -oblj : Add code to start timers here. }
  for I := 0 to Ward.Count - 1 do
    Ward[I].Start;
end;

procedure TScenario.Stop;
var
  I: Integer;
begin
{ DONE 1 -oblj : Add code to stop timers here. }
  for I := 0 to Ward.Count - 1 do
    Ward[I].Stop;
end;

{ TTimedObservationList }

function TTimedObservationList.Add: TTimedObservationItem;
begin
  result := TTimedObservationItem(inherited Add);
end;

procedure TTimedObservationList.Assign(Source: TPersistent);
var
  Obs: TTimedObservationList;
  Item: TTimedObservationItem;
  I: Integer;
begin
  if Source is TTimedObservationList then
  begin
    Clear;
    Obs := TTimedObservationList(Source);
    for I := 0 to Count - 1 do
    begin
      Item := Add;
      Item.Assign(Obs[I]);
    end;
  end
  else
    inherited;
end;

function TTimedObservationList.FindItemID(
  ID: Integer): TTimedObservationItem;
begin
  result := TTimedObservationItem(Inherited FindItemID(ID));
end;

function TTimedObservationList.GetItem(
  Index: Integer): TTimedObservationItem;
begin
  result := TTimedObservationItem(inherited GetItem(Index));
end;

function TTimedObservationList.Insert(
  Index: Integer): TTimedObservationItem;
begin
  result := TTimedObservationItem(inherited Insert(Index));
end;

procedure TTimedObservationList.SetItem(Index: Integer;
  Value: TTimedObservationItem);
begin
  inherited SetItem(Index, Value);
end;



{ TTimedObservationItem }

constructor TTimedObservationItem.Create(Collection: TCollection);
begin
  inherited;
  FTimer := TTimer.Create(nil);
  FTimer.OnTimer := DoTimer;
  FTimer.Enabled := false;
end;

destructor TTimedObservationItem.Destroy;
begin
  FreeAndNil(FTimer);
  inherited;
end;

procedure TTimedObservationItem.DoTimer(Sender: TObject);
begin
  if Assigned(FOnObservation) then
    FOnObservation(Self);
end;

function TTimedObservationItem.GetInterval: Integer;
begin
  result := FTimer.Interval;
end;

procedure TTimedObservationItem.SetInterval(const Value: Integer);
begin
  FTimer.Interval := Value;
end;

end.
 