unit uServers;

interface

uses
  Classes, SysUtils;

type

  EHL7ServerError = class(Exception);

  TServer = class(TCollectionItem)
  private
    FServerName: string;
    FServerIP: string;
    FServerPort: Integer;
  protected
    function GetDisplayName: string;override;
  public
  published
    property ServerName: string read FServerName write FServerName;
    property ServerIP: string read FServerIP write FServerIP;
    property ServerPort: Integer read FServerPort write FServerPort;
    property DisplayName: string read GetDisplayName;
  end;

  TServerList = class(TCollection)
  protected
    function GetItem(Index: Integer): TServer;
    procedure SetItem(Index: Integer; Value: TServer);
  public
    function Add: TServer;
    function Insert(Index: Integer): TServer;
    procedure Load(AFilename: string);
    procedure Save(AFilename: string);
    property Items[Index: Integer]: TServer read GetItem write SetItem; default;
  end;

implementation

uses
  XMLIntf, XMLDoc;

const
  SERVERS = 'SERVERS';
  SERVER = 'SERVER';
  SERVER_NAME = 'SERVER_NAME';
  SERVER_IP = 'SERVER_IP';
  SERVER_PORT = 'SERVER_PORT';

{ TServer }

function TServer.GetDisplayName: string;
begin
  result := Format('%s (%d)', [FServerName, FServerPort]);
end;

{ TServerCollection }

function TServerList.Add: TServer;
begin
  result := TServer(inherited Add);
end;

function TServerList.GetItem(Index: Integer): TServer;
begin
  result := TServer(inherited GetItem(Index));
end;

function TServerList.Insert(Index: Integer): TServer;
begin
  result := TServer(inherited Insert(Index));
end;

procedure TServerList.Load(AFilename: string);
var
  ServerFile: IXMLDocument;
  RootNode: IXMLNode;
  ServerNodes: IXMLNodeList;
  I: Integer;
begin
  ServerFile := LoadXMLDocument(AFilename);
  try
    RootNode := ServerFile.DocumentElement;
    if RootNode.NodeName <> SERVERS then
      raise EHL7ServerError.CreateFmt('Expected server root node to be %s but found %s',
        [SERVERS, RootNode.NodeName]);
    ServerNodes := RootNode.ChildNodes;
    for I := 0 to ServerNodes.Count - 1 do
      with Add do
      begin

        FServerIP := ServerNodes[I].ChildValues[SERVER_IP];
        FServerName := ServerNodes[I].ChildValues[SERVER_NAME];
        FServerPort := ServerNodes[I].ChildValues[SERVER_PORT];
      end;
  finally
    ServerFile := nil;
  end;
end;

procedure TServerList.Save(AFilename: string);
var
  ServerFile: IXMLDocument;
  ServerNode: IXMLNode;
  RootNode: IXMLNode;
  I: Integer;
begin
  ServerFile := NewXMLDocument;
  try
    RootNode := ServerFile.AddChild(SERVERS);

    for I := 0 to Count - 1 do
    begin
      ServerNode := RootNode.AddChild(SERVER);
      ServerNode[SERVER_NAME] := Items[I].FServerName;
      ServerNode[SERVER_IP] := Items[I].FServerIP;
      ServerNode[SERVER_PORT] := Items[I].FServerPort;
    end;
    ServerFile.SaveToFile(AFilename);
  finally
    ServerFile := nil;
  end;
end;

procedure TServerList.SetItem(Index: Integer; Value: TServer);
begin
  inherited SetItem(Index, Value);
end;

end.
